from __future__ import annotations
import copy
import os
import enum
import json
import jsonpatch
import asyncio
from fastapi import Request
from supervisely.app.fastapi.websocket import WebsocketManager
from supervisely.io.fs import dir_exists, mkdir
from supervisely.sly_logger import logger
from supervisely.app.singleton import Singleton
from supervisely.app.fastapi import run_sync


class Field(str, enum.Enum):
    STATE = "state"
    DATA = "data"
    CONTEXT = "context"


def get_data_dir():
    dir = None

    task_id = os.environ.get("TASK_ID")
    if task_id is not None:
        dir = f"/sessions/{task_id}"

    keys = ["SLY_APP_DATA_DIR", "DEBUG_APP_DIR"]
    for key in keys:
        value = os.environ.get(key)
        if value is not None:
            dir = value
            logger.debug(f"Load dir from evn {key}={value}")
            break
    if dir is None:
        raise ValueError(
            f"One of the env variables have to be defined: {[*keys, 'TASK_ID']}"
        )

    if dir_exists(dir) is False:
        logger.info(
            f"App data directory {dir} doesn't exist. Will be made automatically."
        )
        mkdir(dir)
    return dir


class _PatchableJson(dict):
    def __init__(self, field: Field, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._ws = WebsocketManager()
        self._last = copy.deepcopy(dict(self))
        self._lock = asyncio.Lock()
        self._field = field.value

    def get_changes(self, patch=None):
        if patch is None:
            patch = self._get_patch()
        return {self._field: json.loads(patch.to_string())}

    def _get_patch(self):
        patch = jsonpatch.JsonPatch.from_diff(self._last, self)
        return patch

    async def _apply_patch(self, patch):
        async with self._lock:
            patch.apply(self._last, in_place=True)
            self._last = copy.deepcopy(self._last)

    async def synchronize_changes(self):
        patch = self._get_patch()
        await self._apply_patch(patch)
        await self._ws.broadcast(self.get_changes(patch))

    async def send_changes_async(self):
        await self.synchronize_changes()

    def send_changes(self):
        run_sync(self.synchronize_changes())

    def raise_for_key(self, key: str):
        if key in self:
            raise KeyError(f"Key {key} already exists in {self._field}")


class StateJson(_PatchableJson, metaclass=Singleton):
    _global_lock: asyncio.Lock = None

    def __init__(self, *args, **kwargs):
        if StateJson._global_lock is None:
            StateJson._global_lock = asyncio.Lock()
        super().__init__(Field.STATE, *args, **kwargs)

    async def _apply_patch(self, patch):
        await super()._apply_patch(patch)
        # @TODO: _replace_global to patching for optimization
        await StateJson._replace_global(dict(self))

    @classmethod
    async def from_request(cls, request: Request) -> StateJson:
        if "application/json" not in request.headers.get("Content-Type", ""):
            return None
        content = await request.json()
        d = content.get(Field.STATE, {})
        await cls._replace_global(d)
        return cls(d, __local__=True)

    @classmethod
    async def _replace_global(cls, d: dict):
        async with cls._global_lock:
            global_state = cls()
            global_state.clear()
            global_state.update(copy.deepcopy(d))


class DataJson(_PatchableJson, metaclass=Singleton):
    def __init__(self, *args, **kwargs):
        super().__init__(Field.DATA, *args, **kwargs)



# coding: utf-8
from __future__ import annotations
from typing import List, Dict, Optional

from supervisely.video_annotation.video_object import VideoObject
from supervisely.collection.key_indexed_collection import KeyIndexedCollection
from supervisely.project.project_meta import ProjectMeta
from supervisely.video_annotation.key_id_map import KeyIdMap


class VideoObjectCollection(KeyIndexedCollection):
    """
    Collection with :class:`VideoObject<supervisely.video_annotation.video_object.VideoObject>` instances. :class:`VideoObjectCollection<VideoObjectCollection>` object is immutable.

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        # Create two VideoObjects for collection
        class_car = sly.ObjClass('car', sly.Rectangle)
        obj_car = sly.VideoObject(class_car)
        class_bus = sly.ObjClass('bus', sly.Rectangle)
        obj_bus = sly.VideoObject(class_bus)

        # Create VideoObjectCollection
        obj_collection = sly.VideoObjectCollection([obj_car, obj_bus])
        obj_collection_json = obj_collection.to_json()
        print(obj_collection_json)
        # Output: [
        #     {
        #         "key": "773300c59fde4707887068d555269ba5",
        #         "classTitle": "car",
        #         "tags": []
        #     },
        #     {
        #         "key": "8257371497d2402cadabc690f796b1d1",
        #         "classTitle": "bus",
        #         "tags": []
        #     }
        # ]

        # Add item to VideoObjectCollection
        class_truck = sly.ObjClass('truck', sly.Rectangle)
        obj_truck = sly.VideoObject(class_truck)
        # Remember that VideoObjectCollection is immutable, and we need to assign new instance of VideoObjectCollection to a new variable
        new_obj_collection = obj_collection.add(obj_truck)
        new_obj_collection_json = new_obj_collection.to_json()
        print(new_obj_collection_json)
        # Output: [
        #     {
        #         "key": "1b62882f180f49ae96744e13fba26d02",
        #         "classTitle": "car",
        #         "tags": []
        #     },
        #     {
        #         "key": "0c95107fce6c4ac1b42132e83c1a6b3b",
        #         "classTitle": "bus",
        #         "tags": []
        #     },
        #     {
        #         "key": "660d8572e20c4101b4cc7edf7f04b090",
        #         "classTitle": "truck",
        #         "tags": []
        #     }
        # ]

        # You can also add multiple items to collection
        class_truck = sly.ObjClass('truck', sly.Rectangle)
        obj_truck = sly.VideoObject(class_truck)
        class_train = sly.ObjClass('train', sly.Rectangle)
        obj_train = sly.VideoObject(class_train)
        # Remember that VideoObjectCollection is immutable, and we need to assign new instance of VideoObjectCollection to a new variable
        new_obj_collection = obj_collection.add_items([obj_truck, obj_train])
        new_obj_collection_json = new_obj_collection.to_json()
        print(new_obj_collection_json)
        # Output: [
        #     {
        #         "key": "892c62aae48b495687fe8b33ce8ebe96",
        #         "classTitle": "car",
        #         "tags": []
        #     },
        #     {
        #         "key": "db0f64a7b60447769374943077e57679",
        #         "classTitle": "bus",
        #         "tags": []
        #     },
        #     {
        #         "key": "72c24a107f344ef68d4fe8e93dff8184",
        #         "classTitle": "truck",
        #         "tags": []
        #     },
        #     {
        #         "key": "d20738a608234152b43b1c23b7958c47",
        #         "classTitle": "train",
        #         "tags": []
        #     }
        # ]

        # Intersection, finds intersection of given list of instances with collection items

        obj_intersections = obj_collection.intersection([obj_car])
        obj_intersections_json = obj_intersections.to_json()
        print(obj_intersections_json)
        # Output: [
        #     {
        #         "key": "5fae14a4a42c42a29904a887442162c9",
        #         "classTitle": "car",
        #         "tags": []
        #     }
        # ]

        # Difference, finds difference between collection and given list of VideoObject

        obj_diff = obj_collection.difference([obj_car])
        obj_diff_json = obj_diff.to_json()
        print(obj_diff_json)
        # Output: [
        #     {
        #         "key": "b4365bbec0314de58e20267735a39164",
        #         "classTitle": "bus",
        #         "tags": []
        #     }
        # ]

        # Merge, merges collection and given list of VideoObjectCollection

        class_truck = sly.ObjClass('truck', sly.Rectangle)
        obj_truck = sly.VideoObject(class_truck)
        class_train = sly.ObjClass('train', sly.Rectangle)
        obj_train = sly.VideoObject(class_train)
        over_collection = sly.VideoObjectCollection([obj_truck, obj_train])
        # Merge
        merge_collection = obj_collection.merge(over_collection)
        merge_collection_json = merge_collection.to_json()
        print(merge_collection_json)
        # Output: [
        #     {
        #         "key": "8e6f9f05f0954193bdb2005b9cd6e20b",
        #         "classTitle": "truck",
        #         "tags": []
        #     },
        #     {
        #         "key": "a010f91a5e7646bdb8e06dfd6fa50f30",
        #         "classTitle": "train",
        #         "tags": []
        #     },
        #     {
        #         "key": "8440e915a94640b0be279c4fa293553b",
        #         "classTitle": "car",
        #         "tags": []
        #     },
        #     {
        #         "key": "e03e7dd2a2d14854945b6ee7bf7c72e3",
        #         "classTitle": "bus",
        #         "tags": []
        #     }
        # ]
    """
    item_type = VideoObject

    def to_json(self, key_id_map: Optional[KeyIdMap]=None) -> List[Dict]:
        """
        Convert the VideoObjectCollection to a list of json dicts. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :return: List of dicts in json format
        :rtype: :class:`List[dict]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            class_car = sly.ObjClass('car', sly.Rectangle)
            obj_car = sly.VideoObject(class_car)
            class_bus = sly.ObjClass('bus', sly.Rectangle)
            obj_bus = sly.VideoObject(class_bus)
            obj_collection = sly.VideoObjectCollection([obj_car, obj_bus])
            obj_collection_json = obj_collection.to_json()
            print(obj_collection_json)
            # Output: [
            #     {
            #         "key": "773300c59fde4707887068d555269ba5",
            #         "classTitle": "car",
            #         "tags": []
            #     },
            #     {
            #         "key": "8257371497d2402cadabc690f796b1d1",
            #         "classTitle": "bus",
            #         "tags": []
            #     }
            # ]
        """
        return [item.to_json(key_id_map) for item in self]

    @classmethod
    def from_json(cls, data: List[Dict], project_meta: ProjectMeta, key_id_map: Optional[KeyIdMap]=None) -> VideoObjectCollection:
        """
        Convert a list of json dicts to VideoObjectCollection. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :param data: List with dicts in json format.
        :type data: List[dict]
        :param project_meta: Input :class:`ProjectMeta<supervisely.project.project_meta.ProjectMeta>`.
        :type project_meta: ProjectMeta
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :return: VideoObjectCollection object
        :rtype: :class:`VideoObjectCollection`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            obj_collection_json = [
                {
                    "classTitle": "car",
                    "tags": []
                },
                {
                    "classTitle": "bus",
                    "tags": []
                }
            ]

            class_car = sly.ObjClass('car', sly.Rectangle)
            class_bus = sly.ObjClass('bus', sly.Rectangle)
            classes = sly.ObjClassCollection([class_car, class_bus])
            meta = sly.ProjectMeta(obj_classes=classes)

            video_obj_collection = sly.VideoObjectCollection.from_json(obj_collection_json, meta)
        """
        objects = [cls.item_type.from_json(video_object_json, project_meta, key_id_map) for video_object_json in data]
        return cls(objects)

    def __str__(self):
        return 'Objects:\n' + super(VideoObjectCollection, self).__str__()


from supervisely.app.widgets.widget import Widget
from supervisely.app.widgets.radio_table.radio_table import RadioTable
from supervisely.app.widgets.notification_box.notification_box import NotificationBox
from supervisely.app.widgets.done_label.done_label import DoneLabel
from supervisely.app.widgets.sly_tqdm.sly_tqdm import SlyTqdm, Progress
from supervisely.app.widgets.grid_gallery.grid_gallery import GridGallery
from supervisely.app.widgets.classic_table.classic_table import ClassicTable
from supervisely.app.widgets.confusion_matrix.confusion_matrix import ConfusionMatrix
from supervisely.app.widgets.project_selector.project_selector import ProjectSelector
from supervisely.app.widgets.element_button.element_button import ElementButton
from supervisely.app.widgets.button.button import Button
from supervisely.app.widgets.project_thumbnail.project_thumbnail import ProjectThumbnail
from supervisely.app.widgets.apexchart.apexchart import Apexchart
from supervisely.app.widgets.line_chart.line_chart import LineChart
from supervisely.app.widgets.heatmap_chart.heatmap_chart import HeatmapChart
from supervisely.app.widgets.table.table import Table
from supervisely.app.widgets.labeled_image.labeled_image import LabeledImage
from supervisely.app.widgets.text.text import Text

# from supervisely.app.widgets.restart_step.restart_step import RestartStep



# coding: utf-8
"""list/create teams and monitor their activity"""

# docs
from __future__ import annotations
from typing import NamedTuple, List, Dict, Optional, Callable
from supervisely.task.progress import Progress

from supervisely.api.module_api import ApiField, ModuleNoParent, UpdateableModule
from supervisely.sly_logger import logger


# @TODO - umar will add meta with review status and duration
class ActivityAction:
    """
    List of Team Actions to sort Team Activity.
    """

    LOGIN = 'login'
    """"""
    LOGOUT = 'logout'
    """"""
    CREATE_PROJECT = 'create_project'
    """"""
    UPDATE_PROJECT = 'update_project'
    """"""
    DISABLE_PROJECT = 'disable_project'
    """"""
    RESTORE_PROJECT = 'restore_project'
    """"""
    CREATE_DATASET = 'create_dataset'
    """"""
    UPDATE_DATASET = 'update_dataset'
    """"""
    DISABLE_DATASET = 'disable_dataset'
    """"""
    RESTORE_DATASET = 'restore_dataset'
    """"""
    CREATE_IMAGE = 'create_image'
    """"""
    UPDATE_IMAGE = 'update_image'
    """"""
    DISABLE_IMAGE = 'disable_image'
    """"""
    RESTORE_IMAGE = 'restore_image'
    """"""
    CREATE_FIGURE = 'create_figure'
    """"""
    UPDATE_FIGURE = 'update_figure'
    """"""
    DISABLE_FIGURE = 'disable_figure'
    """"""
    RESTORE_FIGURE = 'restore_figure'
    """"""
    CREATE_CLASS = 'create_class'
    """"""
    UPDATE_CLASS = 'update_class'
    """"""
    DISABLE_CLASS = 'disable_class'
    """"""
    RESTORE_CLASS = 'restore_class'
    """"""
    CREATE_BACKUP = 'create_backup'
    """"""
    EXPORT_PROJECT = 'export_project'
    """"""
    MODEL_TRAIN = 'model_train'
    """"""
    MODEL_INFERENCE = 'model_inference'
    """"""
    CREATE_PLUGIN = 'create_plugin'
    """"""
    DISABLE_PLUGIN = 'disable_plugin'
    """"""
    RESTORE_PLUGIN = 'restore_plugin'
    """"""
    CREATE_NODE = 'create_node'
    """"""
    DISABLE_NODE = 'disable_node'
    """"""
    RESTORE_NODE = 'restore_node'
    """"""
    CREATE_WORKSPACE = 'create_workspace'
    """"""
    DISABLE_WORKSPACE = 'disable_workspace'
    """"""
    RESTORE_WORKSPACE = 'restore_workspace'
    """"""
    CREATE_MODEL = 'create_model'
    """"""
    DISABLE_MODEL = 'disable_model'
    """"""
    RESTORE_MODEL = 'restore_model'
    """"""
    ADD_MEMBER = 'add_member'
    """"""
    REMOVE_MEMBER = 'remove_member'
    """"""
    LOGIN_TO_TEAM = 'login_to_team'
    """"""
    ATTACH_TAG = 'attach_tag'
    """"""
    UPDATE_TAG_VALUE = 'update_tag_value'
    """"""
    DETACH_TAG = 'detach_tag'
    """"""
    ANNOTATION_DURATION = 'annotation_duration'
    """"""
    IMAGE_REVIEW_STATUS_UPDATED = 'image_review_status_updated'
    """"""

    # case #1 - labeler pressed "finish image" button in labeling job
    # action: IMAGE_REVIEW_STATUS_UPDATED -> meta["reviewStatus"] == 'done'

    # case #2 - reviewer pressed "accept" or "reject" button
    # action: IMAGE_REVIEW_STATUS_UPDATED -> meta["reviewStatus"] == 'accepted' or 'rejected'

    # possible review statuses:
    # 'done' - i.e. labeler finished the image,
    # 'accepted' - reviewer
    # 'rejected' - reviewer

    # case #3 duration
    # action: ANNOTATION_DURATION -> meta["duration"] e.g. meta-> {"duration": 30} in seconds


class TeamInfo(NamedTuple):
    """
    """
    id: int
    name: str
    description: str
    role: str
    created_at: str
    updated_at: str


class TeamApi(ModuleNoParent, UpdateableModule):
    """
    API for working with Team. :class:`TeamApi<TeamApi>` object is immutable.

    :param api: API connection to the server
    :type api: Api
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        # You can connect to API directly
        address = 'https://app.supervise.ly/'
        token = 'Your Supervisely API Token'
        api = sly.Api(address, token)

        # Or you can use API from environment
        os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
        os.environ['API_TOKEN'] = 'Your Supervisely API Token'
        api = sly.Api.from_env()

        team_info = api.team.get_info_by_id(team_id) # api usage example
    """

    @staticmethod
    def info_sequence():
        """
        NamedTuple TeamInfo containing information about Team.

        :Example:

         .. code-block:: python

            TeamInfo(id=1,
                     name='Vehicle',
                     description='',
                     role='admin',
                     created_at='2020-03-31T14:49:08.931Z',
                     updated_at='2020-03-31T14:49:08.931Z')
        """
        return [
            ApiField.ID,
            ApiField.NAME,
            ApiField.DESCRIPTION,
            ApiField.ROLE,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT
        ]

    @staticmethod
    def info_tuple_name():
        """
        NamedTuple name - **TeamInfo**.
        """
        return 'TeamInfo'

    def __init__(self, api):
        ModuleNoParent.__init__(self, api)
        UpdateableModule.__init__(self, api)

    def get_list(self, filters: List[Dict[str, str]] = None) -> List[TeamInfo]:
        """
        List of all Teams.

        :param filters: List of params to sort output Teams.
        :type filters: list, optional
        :return: List of all Teams with information. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[TeamInfo]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            team_id = 8

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            team_list = api.team.get_list(team_id)
            print(team_list)
            # Output: [TeamInfo(id=1,
            #                   name='Vehicle',
            #                   description='',
            #                   role='admin',
            #                   created_at='2020-03-31T14:49:08.931Z',
            #                   updated_at='2020-03-31T14:49:08.931Z'),
            # TeamInfo(id=2,
            #          name='Road',
            #          description='',
            #          role='admin',
            #          created_at='2020-03-31T08:52:11.000Z',
            #          updated_at='2020-03-31T08:52:11.000Z'),
            # TeamInfo(id=3,
            #          name='Animal',
            #          description='',
            #          role='admin',
            #          created_at='2020-04-02T08:59:03.717Z',
            #          updated_at='2020-04-02T08:59:03.717Z')
            # ]

            # Filtered Team list
            team_list = api.team.get_list(team_id, filters=[{ 'field': 'name', 'operator': '=', 'value': 'Animal' }])
            print(team_list)
            # Output: [TeamInfo(id=3,
            #                  name='Animal',
            #                  description='',
            #                  role='admin',
            #                  created_at='2020-04-02T08:59:03.717Z',
            #                  updated_at='2020-04-02T08:59:03.717Z')
            # ]
        """
        return self.get_list_all_pages('teams.list', {ApiField.FILTER: filters or []})

    def get_info_by_id(self, id: int) -> TeamInfo:
        """
        Get Team information by ID.

        :param id: Team ID in Supervisely.
        :type id: int
        :return: Information about Team. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`TeamInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            team_info = api.team.get_info_by_id(8)
            print(team_info)
            # Output: TeamInfo(id=8,
            #          name='Fruits',
            #          description='',
            #          role='admin',
            #          created_at='2020-04-15T10:50:41.926Z',
            #          updated_at='2020-04-15T10:50:41.926Z')

            # You can also get Team info by name
            team_info = api.team.get_info_by_name("Fruits")
            print(team_info)
            # Output: TeamInfo(id=8,
            #          name='Fruits',
            #          description='',
            #          role='admin',
            #          created_at='2020-04-15T10:50:41.926Z',
            #          updated_at='2020-04-15T10:50:41.926Z')
        """
        return self._get_info_by_id(id, 'teams.info')

    def create(self, name: str, description: Optional[str] = "",
               change_name_if_conflict: Optional[bool] = False) -> TeamInfo:
        """
        Creates Team with given name.

        :param name: Team name.
        :type name: str
        :param description: Team description.
        :type description: str
        :param change_name_if_conflict: Checks if given name already exists and adds suffix to the end of the name.
        :type change_name_if_conflict: bool, optional
        :return: Information about Team. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`TeamInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            new_team = api.team.create("Flowers")
            print(new_team)
            # Output: TeamInfo(id=228,
            #                  name='Flowers',
            #                  description='',
            #                  role='admin',
            #                  created_at='2021-03-11T11:18:46.576Z',
            #                  updated_at='2021-03-11T11:18:46.576Z')
        """
        effective_name = self._get_effective_new_name(name=name, change_name_if_conflict=change_name_if_conflict)
        response = self._api.post('teams.add', {ApiField.NAME: effective_name, ApiField.DESCRIPTION: description})
        return self._convert_json_info(response.json())

    def _get_update_method(self):
        """
        """
        return 'teams.editInfo'

    def get_activity(self, team_id: int,
                     filter_user_id: Optional[int] = None, filter_project_id: Optional[int] = None,
                     filter_job_id: Optional[int] = None,
                     filter_actions: Optional[List] = None, progress_cb: Optional[Callable] = None,
                     start_date: Optional[str] = None,
                     end_date: Optional[str] = None) -> List[Dict]:
        """
        Get Team activity by ID.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param filter_user_id: User ID by which the activity will be filtered.
        :type filter_user_id: int, optional
        :param filter_project_id: Project ID by which the activity will be filtered.
        :type filter_project_id: int, optional
        :param filter_job_id: Job ID by which the activity will be filtered.
        :type filter_job_id: int, optional
        :param filter_actions: List of ActivityAction by which the activity will be filtered.
        :type filter_actions: list, optional
        :param progress_cb: Function to check progress.
        :type progress_cb: Function, optional
        :param start_date: Start date to get Team activity.
        :type start_date: str, optional
        :param end_date: End date to get Team activity.
        :type end_date: str, optional
        :return: Team activity
        :rtype: :class:`List[dict]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            from supervisely.api.team_api import ActivityAction as aa

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            labeling_actions = [
                aa.ATTACH_TAG,
                aa.UPDATE_TAG_VALUE,
                aa.DETACH_TAG,
            ]

            team_activity = api.team.get_activity(8, filter_actions=labeling_actions)
            print(team_activity)
            # Output: [
            #     {
            #         "userId":7,
            #         "action":"detach_tag",
            #         "date":"2021-01-15T15:11:55.985Z",
            #         "user":"cxnt",
            #         "projectId":1817,
            #         "project":"App_Test_Poly",
            #         "datasetId":2370,
            #         "dataset":"train",
            #         "imageId":726985,
            #         "image":"IMG_8144.jpeg",
            #         "classId":"None",
            #         "class":"None",
            #         "figureId":"None",
            #         "job":"None",
            #         "jobId":"None",
            #         "tag":"hhhlk",
            #         "tagId":4720,
            #         "meta":{}
            #     },
            #     {
            #         "userId":7,
            #         "action":"attach_tag",
            #         "date":"2021-01-15T14:24:58.480Z",
            #         "user":"cxnt",
            #         "projectId":1817,
            #         "project":"App_Test_Poly",
            #         "datasetId":2370,
            #         "dataset":"train",
            #         "imageId":726985,
            #         "image":"IMG_8144.jpeg",
            #         "classId":"None",
            #         "class":"None",
            #         "figureId":"None",
            #         "job":"None",
            #         "jobId":"None",
            #         "tag":"hhhlk",
            #         "tagId":4720,
            #         "meta":{}
            #     }
            # ]
        """
        from datetime import datetime, timedelta

        filters = []
        if filter_user_id is not None:
            filters.append({"field": ApiField.USER_ID, "operator": "=", "value": filter_user_id})
        if filter_project_id is not None:
            filters.append({"field": ApiField.PROJECT_ID, "operator": "=", "value": filter_project_id})
        if filter_job_id is not None:
            filters.append({"field": ApiField.JOB_ID, "operator": "=", "value": filter_job_id})
        if filter_actions is not None:
            if type(filter_actions) is not list:
                raise TypeError(
                    "type(filter_actions) is {!r}. But has to be of type {!r}".format(type(filter_actions), list))
            filters.append({"field": ApiField.TYPE, "operator": "in", "value": filter_actions})

        def _add_dt_filter(filters, dt, op):
            dt_iso = None
            if dt is None:
                return
            if type(dt) is str:
                dt_iso = dt
            elif type(dt) is datetime:
                dt_iso = dt.isoformat()
            else:
                raise TypeError('DT type must be string in ISO8601 format or datetime, not {}'.format(type(dt)))
            filters.append({"field": ApiField.DATE, "operator": op, "value": dt_iso})

        _add_dt_filter(filters, start_date, ">=")
        _add_dt_filter(filters, end_date, "<=")

        method = 'teams.activity'
        data = {ApiField.TEAM_ID: team_id, ApiField.FILTER: filters}
        first_response = self._api.post(method, data)
        first_response = first_response.json()

        total = first_response['total']
        per_page = first_response['perPage']
        pages_count = first_response['pagesCount']
        results = first_response['entities']

        if progress_cb is not None:
            progress_cb(len(results), total)
        if pages_count == 1 and len(first_response['entities']) == total:
            pass
        else:
            for page_idx in range(2, pages_count + 1):
                temp_resp = self._api.post(method, {**data, 'page': page_idx, 'per_page': per_page})
                temp_items = temp_resp.json()['entities']
                results.extend(temp_items)
                if progress_cb is not None:
                    progress_cb(len(results), total)
            if len(results) != total:
                logger.warn(f"Method '{method}': new events were created during pagination, "
                            f"downloaded={len(results)}, total={total}")

        return results

    def _convert_json_info(self, info: dict, skip_missing=True) -> TeamInfo:
        """
        """
        res = super()._convert_json_info(info, skip_missing=skip_missing)
        return TeamInfo(**res._asdict())



# coding: utf-8
from __future__ import annotations
from typing import List, Tuple, Dict, Optional
from copy import deepcopy
import uuid
from uuid import UUID
from supervisely.project.project_meta import ProjectMeta


from supervisely._utils import take_with_default
from supervisely.video_annotation.video_tag_collection import VideoTagCollection
from supervisely.video_annotation.video_object_collection import VideoObjectCollection
from supervisely.video_annotation.frame_collection import FrameCollection
from supervisely.video_annotation.constants import FRAMES, IMG_SIZE, IMG_SIZE_HEIGHT, IMG_SIZE_WIDTH, \
                                                       DESCRIPTION, FRAMES_COUNT, TAGS, OBJECTS, VIDEO_ID, KEY, \
                                                       VIDEOS_MAP, VIDEO_NAME
from supervisely.video_annotation.key_id_map import KeyIdMap


class VideoAnnotation:
    """
    VideoAnnotation for a single video. :class:`VideoAnnotation<VideoAnnotation>` object is immutable.

    :param img_size: Size of the image (height, width).
    :type img_size: Tuple[int, int] or List[int, int]
    :param frames_count: Number of frames in VideoAnnotation.
    :type frames_count: int
    :param objects: VideoObjectCollection object.
    :type objects: VideoObjectCollection, optional
    :param frames: FrameCollection object.
    :type frames: FrameCollection, optional
    :param tags: VideoTagCollection object.
    :type tags: VideoTagCollection, optional
    :param description: Video description.
    :type description: str, optional
    :param key: UUID object.
    :type key: UUID, optional
    :raises: :class:`TypeError`, if img_size is not tuple or list
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        # Simple VideoAnnotation example
        height, width = 500, 700
        frames_count = 10
        video_ann = sly.VideoAnnotation((height, width), frames_count)
        print(video_ann.to_json())
        # Output: {
        #     "size": {
        #         "height": 500,
        #         "width": 700
        #     },
        #     "description": "",
        #     "key": "abef780b01ad4063b4b961ab2ba2f410",
        #     "tags": [],
        #     "objects": [],
        #     "frames": [],
        #     "framesCount": 10
        # }

        # More complex VideoAnnotation example

        height, width = 500, 700
        frames_count = 1
        # VideoObjectCollection
        obj_class_car = sly.ObjClass('car', sly.Rectangle)
        video_obj_car = sly.VideoObject(obj_class_car)
        objects = sly.VideoObjectCollection([video_obj_car])
        # FrameCollection
        fr_index = 7
        geometry = sly.Rectangle(0, 0, 100, 100)
        video_figure_car = sly.VideoFigure(video_obj_car, geometry, fr_index)
        frame = sly.Frame(fr_index, figures=[video_figure_car])
        frames = sly.FrameCollection([frame])
        # VideoTagCollection
        meta_car = sly.TagMeta('car_tag', sly.TagValueType.ANY_STRING)
        from supervisely.video_annotation.video_tag import VideoTag
        vid_tag = VideoTag(meta_car, value='acura')
        from supervisely.video_annotation.video_tag_collection import VideoTagCollection
        video_tags = VideoTagCollection([vid_tag])
        # Description
        descr = 'car example'

        video_ann = sly.VideoAnnotation((height, width), frames_count, objects, frames, video_tags, descr)
        print(video_ann.to_json())
        # Output: {
        #     "size": {
        #         "height": 500,
        #         "width": 700
        #     },
        #     "description": "car example",
        #     "key": "a85b282e5e174e7ebad6f878b6919244",
        #     "tags": [
        #         {
        #             "name": "car_tag",
        #             "value": "acura",
        #             "key": "540a8212b0344788953996cea220ea8b"
        #         }
        #     ],
        #     "objects": [
        #         {
        #             "key": "7c74b8a495044ea0ac127f32751c8f5c",
        #             "classTitle": "car",
        #             "tags": []
        #         }
        #     ],
        #     "frames": [
        #         {
        #             "index": 7,
        #             "figures": [
        #                 {
        #                     "key": "82dcbf2e3c5f42a99eeea2ad34173793",
        #                     "objectKey": "7c74b8a495044ea0ac127f32751c8f5c",
        #                     "geometryType": "rectangle",
        #                     "geometry": {
        #                         "points": {
        #                             "exterior": [
        #                                 [
        #                                     0,
        #                                     0
        #                                 ],
        #                                 [
        #                                     100,
        #                                     100
        #                                 ]
        #                             ],
        #                             "interior": []
        #                         }
        #                     }
        #                 }
        #             ]
        #         }
        #     ],
        #     "framesCount": 1
        # }
    """
    def __init__(self, img_size: Tuple[int, int], frames_count: int, objects: Optional[VideoObjectCollection]=None,
                 frames: Optional[FrameCollection]=None, tags: Optional[VideoTagCollection]=None, description: Optional[str]="",
                 key: Optional[UUID]=None):
        if not isinstance(img_size, (tuple, list)):
            raise TypeError('{!r} has to be a tuple or a list. Given type "{}".'.format('img_size', type(img_size)))
        self._img_size = tuple(img_size)
        self._frames_count = frames_count

        self._description = description
        self._tags = take_with_default(tags, VideoTagCollection())
        self._objects = take_with_default(objects, VideoObjectCollection())
        self._frames = take_with_default(frames, FrameCollection())
        self._key = take_with_default(key, uuid.uuid4())

        self.validate_figures_bounds()

    @property
    def img_size(self):
        """
        Size of the image (height, width).

        :return: Image size
        :rtype: :class:`Tuple[int, int]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            height, width = 500, 700
            frames_count = 1
            video_ann = sly.VideoAnnotation((height, width), frames_count)
            print(video_ann.img_size)
            # Output: (500, 700)
        """
        return deepcopy(self._img_size)

    @property
    def frames_count(self):
        """
        Number of frames.

        :return: Frames count
        :rtype: :class:`int`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            height, width = 500, 700
            frames_count = 15
            video_ann = sly.VideoAnnotation((height, width), frames_count)
            print(video_ann.frames_count)
            # Output: 15
        """
        return self._frames_count

    @property
    def objects(self):
        """
        VideoAnnotation objects.

        :return: VideoObjectCollection object
        :rtype: :class:`VideoObjectCollection`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            height, width = 500, 700
            frames_count = 1
            # VideoObjectCollection
            obj_class_car = sly.ObjClass('car', sly.Rectangle)
            video_obj_car = sly.VideoObject(obj_class_car)
            objects = sly.VideoObjectCollection([video_obj_car])
            video_ann = sly.VideoAnnotation((height, width), frames_count, objects)
            print(video_ann.objects.to_json())
            # Output: [
            #     {
            #         "key": "79fc07a4a6ca4b2796279bc033b9ec9a",
            #         "classTitle": "car",
            #         "tags": []
            #     }
            # ]
        """
        return self._objects

    @property
    def frames(self):
        """
        VideoAnnotation frames.

        :return: FrameCollection object
        :rtype: :class:`FrameCollection`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            height, width = 500, 700
            frames_count = 1
            obj_class_car = sly.ObjClass('car', sly.Rectangle)
            video_obj_car = sly.VideoObject(obj_class_car)
            objects = sly.VideoObjectCollection([video_obj_car])
            fr_index = 7
            geometry = sly.Rectangle(0, 0, 100, 100)
            video_figure_car = sly.VideoFigure(video_obj_car, geometry, fr_index)
            frame = sly.Frame(fr_index, figures=[video_figure_car])
            frames = sly.FrameCollection([frame])

            video_ann = sly.VideoAnnotation((height, width), frames_count, objects, frames)
            print(video_ann.frames.to_json())
            # Output: [
            #     {
            #         "index": 7,
            #         "figures": [
            #             {
            #                 "key": "2842f561b1924f6abd6ab6f696ed9b65",
            #                 "objectKey": "7f30fa9b78444ad69e02b37edbf9a902",
            #                 "geometryType": "rectangle",
            #                 "geometry": {
            #                     "points": {
            #                         "exterior": [
            #                             [
            #                                 0,
            #                                 0
            #                             ],
            #                             [
            #                                 100,
            #                                 100
            #                             ]
            #                         ],
            #                         "interior": []
            #                     }
            #                 }
            #             }
            #         ]
            #     }
            # ]
        """
        return self._frames

    @property
    def figures(self):
        """
        VideoAnnotation figures.

        :return: List of VideoFigures from all frames in VideoAnnotation
        :rtype: :class:`List[VideoFigure]<supervisely.video_annotation.video_figure.VideoFigure>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            height, width = 500, 700
            frames_count = 1
            obj_class_car = sly.ObjClass('car', sly.Rectangle)
            video_obj_car = sly.VideoObject(obj_class_car)
            objects = sly.VideoObjectCollection([video_obj_car])
            fr_index = 7
            geometry = sly.Rectangle(0, 0, 100, 100)
            video_figure_car = sly.VideoFigure(video_obj_car, geometry, fr_index)
            frame = sly.Frame(fr_index, figures=[video_figure_car])
            frames = sly.FrameCollection([frame])

            video_ann = sly.VideoAnnotation((height, width), frames_count, objects, frames)
            print(len(video_ann.figures)) # 1
        """
        return self.frames.figures

    @property
    def tags(self):
        """
        VideoAnnotation tags.

        :return: VideoTagCollection object
        :rtype: :class:`VideoTagCollection`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            height, width = 500, 700
            frames_count = 1
            obj_class_car = sly.ObjClass('car', sly.Rectangle)
            video_obj_car = sly.VideoObject(obj_class_car)
            objects = sly.VideoObjectCollection([video_obj_car])
            fr_index = 7
            geometry = sly.Rectangle(0, 0, 100, 100)
            video_figure_car = sly.VideoFigure(video_obj_car, geometry, fr_index)
            frame = sly.Frame(fr_index, figures=[video_figure_car])
            frames = sly.FrameCollection([frame])
            meta_car = sly.TagMeta('car_tag', sly.TagValueType.ANY_STRING)
            from supervisely.video_annotation.video_tag import VideoTag
            vid_tag = VideoTag(meta_car, value='acura')
            from supervisely.video_annotation.video_tag_collection import VideoTagCollection
            tags = VideoTagCollection([vid_tag])

            video_ann = sly.VideoAnnotation((height, width), frames_count, objects, frames, tags)
            print(video_ann.tags.to_json())
            # Output: [
            #     {
            #         "name": "car_tag",
            #         "value": "acura",
            #         "key": "c63e8259589a4fa5b4fb15a48c1f6a63"
            #     }
            # ]
        """
        return self._tags

    def key(self):
        return self._key

    @property
    def description(self):
        """
        Video description.

        :return: Video description
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            height, width = 500, 700
            frames_count = 1
            descr = 'example'
            video_ann = sly.VideoAnnotation((height, width), frames_count, description=descr)
            print(video_ann.description) # example
        """
        return self._description

    def validate_figures_bounds(self) -> None:
        """
        Checks if image contains figures from all frames in collection.

        :raises: :class:`OutOfImageBoundsExtension<supervisely.video_annotation.video_figure.OutOfImageBoundsExtension>`, if figure is out of image bounds
        :return: None
        :rtype: :class:`NoneType`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            height, width = 50, 700
            frames_count = 1
            obj_class_car = sly.ObjClass('car', sly.Rectangle)
            video_obj_car = sly.VideoObject(obj_class_car)
            objects = sly.VideoObjectCollection([video_obj_car])
            fr_index = 7
            geometry = sly.Rectangle(0, 0, 100, 100)
            video_figure_car = sly.VideoFigure(video_obj_car, geometry, fr_index)
            frame = sly.Frame(fr_index, figures=[video_figure_car])
            frames = sly.FrameCollection([frame])

            video_ann = sly.VideoAnnotation((height, width), frames_count, objects, frames)
            video_ann.validate_figures_bounds()
            # raise OutOfImageBoundsExtension("Figure is out of image bounds")
        """
        for frame in self.frames:
            frame.validate_figures_bounds(self.img_size)

    def to_json(self, key_id_map: Optional[KeyIdMap]=None) -> Dict:
        """
        Convert the VideoAnnotation to a json dict. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :return: Json format as a dict
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            height, width = 500, 700
            frames_count = 10
            video_ann = sly.VideoAnnotation((height, width), frames_count)
            print(video_ann.to_json())
            # Output: {
            #     "size": {
            #         "height": 500,
            #         "width": 700
            #     },
            #     "description": "",
            #     "key": "abef780b01ad4063b4b961ab2ba2f410",
            #     "tags": [],
            #     "objects": [],
            #     "frames": [],
            #     "framesCount": 10
            # }
        """
        res_json = {
                        IMG_SIZE: {
                                    IMG_SIZE_HEIGHT: int(self.img_size[0]),
                                    IMG_SIZE_WIDTH: int(self.img_size[1])
                                  },
                        DESCRIPTION: self.description,
                        KEY: self.key().hex,
                        TAGS: self.tags.to_json(key_id_map),
                        OBJECTS: self.objects.to_json(key_id_map),
                        FRAMES: self.frames.to_json(key_id_map),
                        FRAMES_COUNT: self.frames_count
                    }

        if key_id_map is not None:
            video_id = key_id_map.get_video_id(self.key())
            if video_id is not None:
                res_json[VIDEO_ID] = video_id

        return res_json

    @classmethod
    def from_json(cls, data: Dict, project_meta: ProjectMeta, key_id_map: Optional[KeyIdMap]=None) -> VideoAnnotation:
        """
        Convert a json dict to VideoAnnotation. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :param data: Dict in json format.
        :type data: dict
        :param project_meta: Input :class:`ProjectMeta<supervisely.project.project_meta.ProjectMeta>`.
        :type project_meta: ProjectMeta
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :return: VideoAnnotation object
        :rtype: :class:`VideoAnnotation`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            video_ann_json = {
                "size": {
                    "height": 500,
                    "width": 700
                },
                "tags": [],
                "objects": [],
                "frames": [],
                "framesCount": 1
            }
            meta = sly.ProjectMeta()
            video_ann = sly.VideoAnnotation.from_json(video_ann_json, meta)
        """
        #video_name = data[VIDEO_NAME]
        video_key = uuid.UUID(data[KEY]) if KEY in data else uuid.uuid4()

        if key_id_map is not None:
            key_id_map.add_video(video_key, data.get(VIDEO_ID, None))

        img_size_dict = data[IMG_SIZE]
        img_height = img_size_dict[IMG_SIZE_HEIGHT]
        img_width = img_size_dict[IMG_SIZE_WIDTH]
        img_size = (img_height, img_width)

        description = data.get(DESCRIPTION, "")
        frames_count = data[FRAMES_COUNT]

        tags = VideoTagCollection.from_json(data[TAGS], project_meta.tag_metas, key_id_map)
        objects = VideoObjectCollection.from_json(data[OBJECTS], project_meta, key_id_map)
        frames = FrameCollection.from_json(data[FRAMES], objects, frames_count, key_id_map)

        return cls(img_size=img_size,
                   frames_count=frames_count,
                   objects=objects,
                   frames=frames,
                   tags=tags,
                   description=description,
                   key=video_key)

    def clone(self, img_size: Optional[Tuple[int, int]]=None, frames_count: Optional[int]=None,
              objects: Optional[VideoObjectCollection]=None, frames: Optional[FrameCollection]=None,
              tags: Optional[VideoTagCollection]=None, description: Optional[str]=None) -> VideoAnnotation:
        """
        Makes a copy of VideoAnnotation with new fields, if fields are given, otherwise it will use fields of the original VideoAnnotation.

        :param img_size: Size of the image (height, width).
        :type img_size: Tuple[int, int], optional
        :param frames_count: Number of frames in VideoAnnotation.
        :type frames_count: int, optional
        :param objects: VideoObjectCollection object.
        :type objects: VideoObjectCollection, optional
        :param frames: FrameCollection object.
        :type frames: FrameCollection, optional
        :param tags: VideoTagCollection object.
        :type tags: VideoTagCollection, optional
        :param description: Video description.
        :type description: str, optional
        :raises: :class:`TypeError`, if img_size is not tuple or list
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            height, width = 500, 700
            frames_count = 1
            video_ann = sly.VideoAnnotation((height, width), frames_count)

            obj_class_car = sly.ObjClass('car', sly.Rectangle)
            video_obj_car = sly.VideoObject(obj_class_car)
            new_objects = sly.VideoObjectCollection([video_obj_car])
            new_video_ann = video_ann.clone(objects=new_objects)
            print(new_video_ann.to_json())
            # Output: {
            #     "size": {
            #         "height": 500,
            #         "width": 700
            #     },
            #     "description": "",
            #     "key": "37f7d267864c4fd8b1a1a32f67e37f7d",
            #     "tags": [],
            #     "objects": [
            #         {
            #             "key": "27d4ba1aaee64930b2d0bfb7e8b53493",
            #             "classTitle": "car",
            #             "tags": []
            #         }
            #     ],
            #     "frames": [],
            #     "framesCount": 1
            # }
        """
        return VideoAnnotation(img_size=take_with_default(img_size, self.img_size),
                               frames_count=take_with_default(frames_count, self.frames_count),
                               objects=take_with_default(objects, self.objects),
                               frames=take_with_default(frames, self.frames),
                               tags=take_with_default(tags, self.tags),
                               description=take_with_default(description, self.description))


    def is_empty(self) -> bool:
        """
        Check whether video annotation contains objects or tags, or not.

        :returns: True if video annotation is empty, False otherwise.
        :rtype: :class:`bool`
        """
        if len(self.objects) == 0 and len(self.tags) == 0:
            return True
        else:
            return False



# coding: utf-8
"""list available classes in supervisely project"""

# docs
from typing import Optional, List, Dict

from supervisely.api.module_api import ModuleApi
from supervisely.api.module_api import ApiField


class ProjectClassApi(ModuleApi):
    """
    API for working with classes in Project. :class:`ProjectClassApi<ProjectClassApi>` object is immutable.

    :param api: API connection to the server
    :type api: Api
    """
    @staticmethod
    def info_sequence():
        return [ApiField.ID,
                #ApiField.PROJECT_ID,
                ApiField.NAME,
                ApiField.DESCRIPTION,
                ApiField.SHAPE,
                ApiField.COLOR,
                ApiField.CREATED_AT,
                ApiField.UPDATED_AT
                ]

    @staticmethod
    def info_tuple_name():
        """
        NamedTuple name - **ProjectClassInfo**.
        """
        return 'ProjectClassInfo'

    def get_list(self, project_id: int, filters: Optional[List[Dict[str, str]]] = None) -> list:
        """
        List of Classes in the given Project.

        :param project_id: Project ID in Supervisely.
        :type project_id: int
        :param filters:
        :type filters: list
        :return: List of classes.
        :rtype: :class:`list`
        """
        return self.get_list_all_pages('advanced.object_classes.list',  {ApiField.PROJECT_ID: project_id, "filter": filters or []})
